import functools
import operator

from .pdgrid import grid_values

class GridRequest():

    def __init__(self, request):
        self.request = request

    @property
    def aggregation_group(self):
        '''The column to group the data by (first unexpanded groupby column)'''
        if len(self.selected_groups) < len(self.groupby_columns):
            return self.groupby_columns[len(self.selected_groups)]
        return None


class column_filter(functools.partial):
    '''Like partial, but allow to pass in leftmost argument last'''
    def __call__(self, column):
        return self.func(column, *self.args)


def operator_processor(column, operator_func, filter1, filter2):
    '''Combine the output of two column filters using the operator (OR, AND) passed in'''
    return operator_func(filter1(column), filter2(column))

operator_map = {'OR': operator.or_,
                'AND': operator.and_}



class AgGridRequest(GridRequest):

    agg_lookup = {
        'avg': 'mean'
    }

    FILTER_OPERATORS = {
        ('set', None): lambda column, filter_value: column.astype(str).isin(filter_value),
        ('text', 'contains'): lambda column, filter_value: column.str.contains(filter_value),
        ('text', 'notContains'): lambda column, filter_value: ~column.str.contains(filter_value),
        ('text', 'equals'): operator.eq,
        ('text', 'notEquals'): operator.ne,
        ('text', 'startsWith'): lambda column, filter_value: column.str.startswith(filter_value),
        ('text', 'endsWith'): lambda column, filter_value: column.str.endswith(filter_value),
        ('number', 'lessThanOrEqual'): operator.le,
        ('number', 'lessThan'): operator.lt,
        ('number', 'equals'): operator.eq,
        ('number', 'notEquals'): operator.ne,
        ('number', 'greaterThan'): operator.gt,
        ('number', 'greaterThanOrEqual'): operator.ge,
    }


    def filter_to_func(self, f):
        if 'operator' in f:
            return column_filter(operator_processor,
                                 operator_map[f['operator']],
                                 self.filter_to_func(f.get('condition1')),
                                 self.filter_to_func(f.get('condition2'))
                                 )
        elif f.get('filterType') == 'number' and f.get('type') == 'inRange':
            return column_filter(operator_processor,
                                 operator_map['AND'],
                                 column_filter(self.FILTER_OPERATORS[('number', 'greaterThanOrEqual')], f.get('filter')),
                                 column_filter(self.FILTER_OPERATORS[('number', 'lessThan')], f.get('filterTo'))
                                 )
        else:
            return column_filter(self.FILTER_OPERATORS[(f['filterType'], f.get('type'))], f.get('values' if f['filterType'] == 'set' else 'filter'))
        
    @property
    def groupby_columns(self):
        '''The grouped columns in the grid'''
        return [f.get('field') for f in self.request.get('rowGroupCols', [])]

    @property
    def selected_groups(self):
        '''The groups that have been expanded'''
        return self.request.get('groupKeys', [])

    @property
    def aggregation_params(self):
        return {f.get('field'): self.agg_lookup.get(f.get('aggFunc'), f.get('aggFunc'))
                for f in self.request.get('valueCols')
                if f.get('aggFunc') is not None
                and f.get('field') is not None}
            
    @property
    def sort_columns(self):
        return {f.get('colId'):f.get('sort') == 'asc' for f in self.request.get('sortModel', [])
                if f.get('colId') not in self.groupby_columns[:len(self.selected_groups)]}

    @property
    def start_row(self):
        return self.request.get('startRow')

    @property
    def end_row(self):
        return self.request.get('endRow')

    @property
    def filter_model(self):
        return { field: self.filter_to_func(f) for field, f in self.request.get('filterModel', {}).items()}

    @property
    def expanded_column_filters(self):
        return { field: column_filter(operator.eq, filter_value) for field, filter_value in zip(self.groupby_columns, self.selected_groups)}
    
    @property
    def filters(self):
        return {**self.filter_model,  **self.expanded_column_filters}

    @property
    def pivot_mode(self):
        return self.request.get('pivotMode', False)

    @property
    def pivot_columns(self):
        return [f.get('field') for f in self.request.get('pivotCols', [])][:1]

    @property
    def do_pivot(self):
        return self.pivot_mode and self.pivot_columns
    
    @property
    def sort_agg_params(self):
        '''Columns needed for sorting but not being aggregated'''
        return {f: ('min' if asc else 'max') for f, asc in self.sort_columns.items()
                if f not in self.aggregation_params and f != self.aggregation_group}

    @property
    def sort_fields(self):
        return list(self.sort_columns.keys())

    @property
    def sort_ascending(self):
        return list(self.sort_columns.values())

    @property
    def display_names(self):
        return {f['field']: f['displayName'] for f in self.request['valueCols']}



def aggrid_values(df, request):
    '''Entry point for aggrid request'''
    return grid_values(df, AgGridRequest(request))



import pandas as pd
import operator
import functools
import itertools

import numpy as np



def unique_values(df, filter_field):
    '''Endpoint that provides a distinct list of values in a column, used by the front end for filtering'''
    return df[filter_field].drop_duplicates().sort_values().tolist()

def grid_values(df, request):

    response = {'secondaryColDefs': {}}
    df = process_filters(df, request)
    df = sort_and_aggregate(df, request, response)

    response['rows'] = df.to_dict(orient='records')

    return response


def pivot(df, request, response):

    df = df.pivot_table(
        index = request.aggregation_group,
        columns = request.pivot_columns,
        values = list(request.aggregation_params),
        aggfunc=request.aggregation_params)

    if not request.aggregation_group:
        df = df.unstack()
        df = pd.DataFrame(columns=df.index, data=[df.values])
    else:
        df.columns = df.columns.swaplevel(0,1)
    
    secondaryColDefs = [
]
    for group, cols in itertools.groupby(sorted(df.columns, key=lambda k: k[0]), key=lambda k: k[0]):
        secondaryColDefs.append({'groupId': str(group),
                                 'headerName': str(group),
                                 'children': [
                                     {'colId': "|".join(str(f) for f in col),
                                      'field': "|".join(str(f) for f in col),
                                      'headerName': f"{request.aggregation_params[col[-1]]}({request.display_names[col[-1]]})"}
                                     for col in cols]
                                 }
                                )
    df.columns = ["|".join(str(f) for f in col) for col in df.columns] 
    
    df = df.replace({np.nan: None})

    response['secondaryColDefs'] = secondaryColDefs
    return df
    
def sort_and_aggregate(df, request, response):

    if request.aggregation_group:
        df = df.set_index(request.aggregation_group)

        groups = df.groupby(request.aggregation_group)
        last_row = groups.ngroups

        #Do we need to paginate or sort on non aggregated field? If so, paginate before aggregating all columns. Huge speed up in some cases!!
        if pagination := request.end_row < last_row or request.start_row > 0 or request.sort_agg_params:

            if not request.sort_fields or request.sort_fields[0] == request.aggregation_group:
                #only sort by group by column
                ascending = True if not request.sort_fields else request.sort_ascending[0]
                sorted_keys = df.index.drop_duplicates().sort_values(ascending=ascending)
            elif request.do_pivot:
                pivot_sort_column = next(iter(request.sort_agg_params.keys()))
                pre_pivot_column = pivot_sort_column.split('|')[-1]
                sorted_keys = df.pivot_table(index=df.index, columns=request.pivot_columns, values=[pre_pivot_column], aggfunc=request.aggregation_params.get(pre_pivot_column))
                sorted_keys.columns = ["|".join(str(f) for f in col[::-1]) for col in sorted_keys.columns] 
                sorted_keys = sorted_keys.sort_values(by=request.sort_fields, ascending=request.sort_ascending).index

            else:    
                agg_params = {**request.sort_agg_params, **{k:v for k,v in request.aggregation_params.items() if k in request.sort_fields}}
                sorted_keys = (groups
                               .agg(agg_params)
                               .sort_values(by=list(request.sort_columns.keys()),
                                            ascending=list(request.sort_columns.values()))
                               ).index
                
            sorted_keys = sorted_keys[request.start_row:request.end_row]
            df = df.loc[sorted_keys] #Filter out unneded groups before aggregating

        if request.do_pivot:
            df = pivot(df, request, response)
        else:
            df = df.groupby(level=0).agg(request.aggregation_params)
        if pagination:
            df = df.loc[sorted_keys] #Group by scrambles order, resort
        df = df.reset_index(drop=False)

    #Pivot without aggregation group
    elif request.do_pivot:
        df = pivot(df, request, response)
        
    if (not request.aggregation_group or not pagination):
        df = df.sort_values(by=request.sort_fields, ascending=request.sort_ascending)
        last_row = df.index.size
        df = df.iloc[request.start_row:request.end_row]

    response['lastRow'] = last_row
    return df


def process_filters(df, request):
    '''This could possibly be done with df.query, but unclear if it would be faster'''
    for field, filter_function in request.filters.items():
        df = df.loc[filter_function(df[field])]
    return df

